// Copyright 2006-2007 The Parancoe Team
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.parancoe.persistence.dao.generic;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Properties;
import org.apache.log4j.Logger;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.reflect.MethodSignature;
import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.MappingException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.criterion.DetachedCriteria;
import org.hibernate.criterion.Order;
import org.hibernate.criterion.Restrictions;
import org.springframework.orm.hibernate3.HibernateCallback;
import org.springframework.util.StringUtils;

/**
 * An aspect for instrumenting a class with dao interfaces and methods.
 *
 * @author <a href="mailto:lucio.benfante@jugpadova.it">Lucio Benfante</a>
 * @version $Revision: 507854cbcafb $
 */
@Aspect
public class HibernateDaoInstrumentation {

    private static final Logger logger = Logger.getLogger(
            HibernateDaoInstrumentation.class);
    private ExceptionManager exceptionManager;

    public void setExceptionManager(ExceptionManager exceptionManager) {
        this.exceptionManager = exceptionManager;
    }

    @Around(
    value =
    "execution(* *(..)) && target(org.parancoe.persistence.dao.generic.GenericDaoHibernateSupport)")
    public Object executeFinder(ProceedingJoinPoint pjp) throws Throwable {
        Object result = null;
        final GenericDaoHibernateSupport target = (GenericDaoHibernateSupport) pjp.
                getTarget();
        final Method method = ((MethodSignature) pjp.getSignature()).getMethod();
        final Object[] args = pjp.getArgs();
        final Class<?>[] parameterTypes = method.getParameterTypes();
        final Annotation[][] parameterAnnotations = method.
                getParameterAnnotations();
        final Integer firstResult = getFirstResultValue(args, parameterTypes,
                parameterAnnotations);
        final Integer maxResults = getMaxResultsValue(args, parameterTypes,
                parameterAnnotations);
        logger.debug("target: " + target);
        logger.debug("method: " + method);
        logger.debug("args: " + args);

        // query using a named query from the method name
        result = target.getHibernateTemplate().executeFind(new HibernateCallback() {

            @Override
            public Object doInHibernate(Session session) throws
                    HibernateException {
                String queryName = queryNameFromMethod(target, method);
                Query namedQuery = null;
                try {
                    namedQuery = session.getNamedQuery(queryName);
                } catch (MappingException e) {
                    // No such named query
                    logger.debug("Named query not found: "+queryName+".");
                }
                if (namedQuery != null) {
                    for (int i = 0; i < args.length; i++) {
                        if (isQueryParameter(i, parameterTypes,
                                parameterAnnotations)) {
                            Object arg = args[i];
                            namedQuery.setParameter(i, arg);
                        }
                    }
                    if (firstResult != null) {
                        namedQuery.setFirstResult(firstResult.intValue());
                    }
                    if (maxResults != null && maxResults.intValue() >= 0) {
                        namedQuery.setMaxResults(maxResults.intValue());
                    }
                    namedQuery.setCacheable(method.isAnnotationPresent(
                            CacheIt.class));
                    return namedQuery.list();
                }
                return null;
            }
        });
        if (result == null) {
            // No named query found
            if (method.getName().startsWith("findBy")) {
                // Query evicting condition from the method name
                result = target.getHibernateTemplate().executeFind(new HibernateCallback() {

                    @Override
                    public Object doInHibernate(Session session) throws
                            HibernateException {
                        DetachedCriteria criteria = criteriaFromMethod(target,
                                method, args);
                        Criteria executableCriteria = criteria.
                                getExecutableCriteria(session);
                        if (firstResult != null) {
                            executableCriteria.setFirstResult(firstResult.
                                    intValue());
                        }
                        if (maxResults != null && maxResults.intValue() >= 0) {
                            executableCriteria.setMaxResults(
                                    maxResults.intValue());
                        }
                        final Criteria crit = criteria.getExecutableCriteria(
                                session);
                        crit.setCacheable(method.isAnnotationPresent(
                                CacheIt.class));
                        return crit.list();
                    }
                });
            } else {
                // Call an instance method
                try {
                    result = pjp.proceed(args);
                } catch (final Throwable throwable) {
                    final Properties p = new Properties();
                    p.setProperty(ExceptionManager.METHOD_NAME, method.getName());
                    p.setProperty(ExceptionManager.TARGET_CLASS_NAME, target.
                            getClass().getName());
                    exceptionManager.process(throwable, p);
                }
            }
        }
        if (result != null && List.class.isAssignableFrom(result.getClass()) && !List.class.
                isAssignableFrom(method.getReturnType())) {
            // The return type is not a List, so I return the first result
            // of the list, or null if the list is empty
            List listResult = (List) result;
            if (!listResult.isEmpty()) {
                result = listResult.get(0);
            } else {
                result = null;
            }
        }
        return result;
    }

    private String queryNameFromMethod(GenericDaoHibernateSupport target,
            Method finderMethod) {
        return target.getType().getSimpleName() + "." + finderMethod.getName();
    }

    private DetachedCriteria criteriaFromMethod(
            GenericDaoHibernateSupport target, Method finderMethod,
            Object[] args) {
        Class<?>[] parameterTypes = finderMethod.getParameterTypes();
        Annotation[][] parameterAnnotations = finderMethod.
                getParameterAnnotations();
        DetachedCriteria criteria = DetachedCriteria.forClass(target.getType());
        int orderByIdx = finderMethod.getName().indexOf("OrderBy");
        String[] parameters = null;
        String[] orderParameters = null;
        if (orderByIdx == -1) {
            // no orderBy
            parameters = finderMethod.getName().substring(6).split("And");
        } else {
            if (orderByIdx - 1 > 6) {
                parameters = finderMethod.getName().substring(6, orderByIdx).
                        split("And");
            }
            orderParameters = finderMethod.getName().substring(orderByIdx + 7).
                    split("And");
        }
        if (parameters != null) {
            int argIndex = 0;
            for (int i = 0; i < parameters.length; i++) {
                while (!isQueryParameter(argIndex, parameterTypes,
                        parameterAnnotations)) {
                    // skip not query parameters
                    argIndex++;
                }
                addComparison(criteria, StringUtils.uncapitalize(parameters[i]),
                        args[argIndex], parameterAnnotations[argIndex]);
//                criteria.add(Restrictions.eq(StringUtils.uncapitalize(parameters[i]), args[argIndex]));
                argIndex++;
            }
        }
        if (orderParameters != null) {
            for (String oPar : orderParameters) {
                criteria.addOrder(Order.asc(StringUtils.uncapitalize(oPar)));
            }
        }
        return criteria;
    }

    /**
     * Check if a parameter is for the FirstResult value for the query.
     *
     * @param parameterIndex index of the parameter as declared in the method
     * @param parameterTypes types of parameters of the method
     * @param parameterAnnotations annotations of parameters of the method
     * @return true if the parameter is for the FirstResult value
     */
    private boolean isFirstResultParameter(int parameterIndex,
            Class<?>[] parameterTypes, Annotation[][] parameterAnnotations) {
        boolean result = false;
        for (Annotation annotation : parameterAnnotations[parameterIndex]) {
            if (annotation instanceof FirstResult && int.class.isAssignableFrom(
                    parameterTypes[parameterIndex])) {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Check if a parameter is for the MaxResults value for the query.
     *
     * @param parameterIndex index of the parameter as declared in the method
     * @param parameterTypes types of parameters of the method
     * @param parameterAnnotations annotations of parameters of the method
     * @return true if the parameter is for the MaxResults value
     */
    private boolean isMaxResultsParameter(int parameterIndex,
            Class<?>[] parameterTypes, Annotation[][] parameterAnnotations) {
        boolean result = false;
        for (Annotation annotation : parameterAnnotations[parameterIndex]) {
            if (annotation instanceof MaxResults && int.class.isAssignableFrom(
                    parameterTypes[parameterIndex])) {
                result = true;
                break;
            }
        }
        return result;
    }

    /**
     * Check if a parameter is a parameter for the query
     *
     * @param parameterIndex index of the parameter as declared in the method
     * @param parameterTypes types of parameters of the method
     * @param parameterAnnotations annotations of parameters of the method
     * @return true if the parameter is a parameter for the query
     */
    private boolean isQueryParameter(int parameterIndex,
            Class<?>[] parameterTypes, Annotation[][] parameterAnnotations) {
        return !isMaxResultsParameter(parameterIndex, parameterTypes,
                parameterAnnotations) && !isFirstResultParameter(parameterIndex,
                parameterTypes, parameterAnnotations);
    }

    /**
     * Get the value of the parameter marked as @FirstRecord
     *
     * @param args parameters values
     * @param parameterTypes types of parameters of the method
     * @param parameterAnnotations annotations of parameters of the method
     * @return The value if it's found. null otherwise.
     */
    private Integer getFirstResultValue(Object args[], Class<?>[] parameterTypes,
            Annotation[][] parameterAnnotations) {
        Integer result = null;
        for (int i = 0; i < args.length; i++) {
            if (isFirstResultParameter(i, parameterTypes, parameterAnnotations)) {
                result = (Integer) args[i];
                break;
            }
        }
        return result;
    }

    /**
     * Get the value of the parameter marked as @MaxResults
     *
     * @param args parameters values
     * @param parameterTypes types of parameters of the method
     * @param parameterAnnotations annotations of parameters of the method
     * @return The value if it's found. null otherwise.
     */
    private Integer getMaxResultsValue(Object args[], Class<?>[] parameterTypes,
            Annotation[][] parameterAnnotations) {
        Integer result = null;
        for (int i = 0; i < args.length; i++) {
            if (isMaxResultsParameter(i, parameterTypes, parameterAnnotations)) {
                result = (Integer) args[i];
                break;
            }
        }
        return result;
    }

    /**
     * Add a comparison to criteria.
     *
     * @param criteria
     * @param string the parameter name
     * @param object the parameter value to compare
     * @param annotations method parameter annotations
     */
    private void addComparison(DetachedCriteria criteria, String parameter,
            Object value, Annotation[] annotations) {
        Compare compareAnnotation = getCompareAnnotation(annotations);
        CompareType compareType = compareAnnotation != null ? compareAnnotation.
                value() : CompareType.EQUAL;
        switch (compareType) {
            case LIKE:
                criteria.add(Restrictions.like(parameter, value));
                break;
            case ILIKE:
                criteria.add(Restrictions.ilike(parameter, value));
                break;
            case GE:
                criteria.add(Restrictions.ge(parameter, value));
                break;
            case GT:
                criteria.add(Restrictions.gt(parameter, value));
                break;
            case LE:
                criteria.add(Restrictions.le(parameter, value));
                break;
            case LT:
                criteria.add(Restrictions.lt(parameter, value));
                break;
            case NE:
                criteria.add(Restrictions.ne(parameter, value));
                break;
            case EQUAL:
            default:
                criteria.add(Restrictions.eq(parameter, value));
        }
    }

    private Compare getCompareAnnotation(Annotation[] annotations) {
        Compare result = null;
        for (Annotation annotation : annotations) {
            if (annotation instanceof Compare) {
                result = (Compare) annotation;
                break;
            }
        }
        return result;
    }
}
